/*==========================================================================;
 *
 *  Copyright (C) 1995-2001 Microsoft Corporation.  All Rights Reserved.
 *
 *  File:       idirect3ddevice8.cpp
 *  Content:    Direct3D capabilities include file
 *
 ***************************************************************************/
#include "d3dprj.h"
#include "gl_init.h"

static GLenum D3D8_GlEnum(D3DPRIMITIVETYPE type)
{
	switch(type)
	{
		case D3DPT_POINTLIST: return GL_POINTS;
		case D3DPT_LINELIST: return GL_LINE_LOOP;
		case D3DPT_LINESTRIP: return GL_LINE_STRIP;
		case D3DPT_TRIANGLELIST: return GL_TRIANGLES;
		case D3DPT_TRIANGLESTRIP: return GL_TRIANGLE_STRIP;
		case D3DPT_TRIANGLEFAN: return GL_TRIANGLE_FAN;
	}	
	return GL_POLYGON; // ?
}

static int D3D8_GetTexCoordSize(unsigned stage, unsigned fvf)
{
	if (fvf & D3DFVF_TEXCOORDSIZE2(stage))
		return 2;

	if (fvf & D3DFVF_TEXCOORDSIZE1(stage))
		return 1;

	if (fvf & D3DFVF_TEXCOORDSIZE3(stage))
		return 3;

	if (fvf & D3DFVF_TEXCOORDSIZE4(stage))
		return 4;

	return 2;
}

static int D3D8_DecodeLength(unsigned fvf)
{
	int nSize = 0;
	if (fvf & D3DFVF_XYZ)		nSize+=3 * sizeof(FLOAT);
	if (fvf & D3DFVF_XYZRHW)	nSize+=4 * sizeof(FLOAT);
	if (fvf & D3DFVF_NORMAL)	nSize+=3 * sizeof(FLOAT);
	if (fvf & D3DFVF_DIFFUSE)	nSize+=4 * sizeof(BYTE);
	if (fvf & D3DFVF_SPECULAR)	nSize+=4 * sizeof(BYTE);
	if (fvf & D3DFVF_TEX1)		nSize+=D3D8_GetTexCoordSize(0, fvf) * sizeof(FLOAT);
	else
	if (fvf & D3DFVF_TEX2)		nSize+=D3D8_GetTexCoordSize(1, fvf) * sizeof(FLOAT);
	else		
	if (fvf & D3DFVF_TEX3)		nSize+=D3D8_GetTexCoordSize(2, fvf) * sizeof(FLOAT);
	else
	if (fvf & D3DFVF_TEX4)		nSize+=D3D8_GetTexCoordSize(3, fvf) * sizeof(FLOAT);
	else
	if (fvf & D3DFVF_TEX5)		nSize+=D3D8_GetTexCoordSize(4, fvf) * sizeof(FLOAT);
	else
	if (fvf & D3DFVF_TEX6)		nSize+=D3D8_GetTexCoordSize(5, fvf) * sizeof(FLOAT);
	else
	if (fvf & D3DFVF_TEX7)		nSize+=D3D8_GetTexCoordSize(6, fvf) * sizeof(FLOAT);
	else
	if (fvf & D3DFVF_TEX8)		nSize+=D3D8_GetTexCoordSize(7, fvf) * sizeof(FLOAT);
	
	return nSize;
}

int D3D8_DrawIndexedBufferToGL(D3DPRIMITIVETYPE type, unsigned fvf, const void *verticies, int count, const void *indicies, int stride)
{
	int size = 0;
	int fmt = 0;
	int vertex_count = 3 * count;	

	const char *vtx = (const char *) verticies;
	int sstride = stride + D3D8_DecodeLength(fvf);	

	glEnableClientState( GL_VERTEX_ARRAY );		

	if (fvf & D3DFVF_XYZ)	
	{
		glVertexPointer(3, GL_FLOAT, sstride, vtx);
		vtx+=3 * 4;
	}
	
	if (fvf & D3DFVF_XYZRHW)
	{		
		glVertexPointer(4, GL_FLOAT, sstride, vtx);
		float *f = (float*) vtx;
		vtx+=4 * 4;
	}

	if (fvf & D3DFVF_NORMAL)
	{
		glEnableClientState( GL_NORMAL_ARRAY );
		glNormalPointer(GL_FLOAT, sstride, vtx);
		vtx+=3 * 4;
	}
	else
		glDisableClientState( GL_NORMAL_ARRAY );

	
	if (fvf & D3DFVF_DIFFUSE)
	{
		glEnableClientState( GL_COLOR_ARRAY );
		glColorPointer(4, GL_UNSIGNED_BYTE, sstride, vtx);
		vtx+=4;
	}
	else
		glDisableClientState( GL_COLOR_ARRAY );

	if (fvf & D3DFVF_TEX1)
	{
		int n = D3D8_GetTexCoordSize(0, fvf);
		glEnableClientState( GL_TEXTURE_COORD_ARRAY );			
		glTexCoordPointer(n, GL_FLOAT, sstride, vtx);
		vtx+=n * 4;
	}
	else
		glDisableClientState( GL_TEXTURE_COORD_ARRAY );

#ifdef USE_LOCK_ARRAYS	
	if (gl_EXT_lock_arrays)
		glLockArraysEXT(0, vertex_count);
#endif

	glDrawElements( D3D8_GlEnum(type), vertex_count, GL_UNSIGNED_SHORT, indicies );

#ifdef USE_LOCK_ARRAYS	
	if (gl_EXT_lock_arrays)
		glUnlockArraysEXT(); // unlock vertex arrays
#endif

	return 1;

}

const char *D3D8_DecodeToGLInstruction(const char *pInstruction, unsigned fvf)
{	
	const char *p0 = pInstruction;
	if (fvf & D3DFVF_XYZ)
		pInstruction+=3 * sizeof(FLOAT);

	if (fvf & D3DFVF_XYZRHW)
		pInstruction+=4 * sizeof(FLOAT);

	if (fvf & D3DFVF_NORMAL)
	{
		glNormal3fv((GLfloat*)pInstruction);
		pInstruction+=3 * sizeof(FLOAT);
	}
	if (fvf & D3DFVF_DIFFUSE)
	{
		GLubyte *c = (GLubyte*)pInstruction;
		glColor4ubv(c);
		pInstruction+=4 * sizeof(BYTE);		
	}

	if (fvf & D3DFVF_SPECULAR)
	{
		// TODO: use secondary color extension
		pInstruction+=4 * sizeof(BYTE);
	}

	if (fvf & D3DFVF_TEX1)
	{
		glTexCoord3fv((GLfloat*)pInstruction);
		pInstruction+=D3D8_GetTexCoordSize(0, fvf) * sizeof(FLOAT);
	}
	else
	if (fvf & D3DFVF_TEX2)
	{
		// TODO: use ARB multitexture extension
		pInstruction+=D3D8_GetTexCoordSize(1, fvf) * sizeof(FLOAT);
	}
	else
	if (fvf & D3DFVF_TEX3)
	{
		pInstruction+=D3D8_GetTexCoordSize(2, fvf) * sizeof(FLOAT);
	}
	else
	if (fvf & D3DFVF_TEX4)
	{
		pInstruction+=D3D8_GetTexCoordSize(3, fvf) * sizeof(FLOAT);
	}
	else
	if (fvf & D3DFVF_TEX5)
	{
		pInstruction+=D3D8_GetTexCoordSize(4, fvf) * sizeof(FLOAT);
	}
	else
	if (fvf & D3DFVF_TEX6)
	{
		pInstruction+=D3D8_GetTexCoordSize(5, fvf) * sizeof(FLOAT);
	}
	else
	if (fvf & D3DFVF_TEX7)
	{
		pInstruction+=D3D8_GetTexCoordSize(6, fvf) * sizeof(FLOAT);
	}
	else
	if (fvf & D3DFVF_TEX8)
	{
		pInstruction+=D3D8_GetTexCoordSize(7, fvf) * sizeof(FLOAT);
	}

	if (fvf & D3DFVF_XYZ)
	{
		float *v = (float *)p0;
		glVertex3fv(v);
	}

	if (fvf & D3DFVF_XYZRHW)
	{
		float *v = (float *)p0;
		glVertex4fv(v);
	}
	return pInstruction;
}

BIDirect3DDevice8::BIDirect3DDevice8(IDirect3D8 *p) : m_refCount(0)
{
	m_pIDirect3D8 = p;

	memset(m_RenderState,0, 256 * 4);	
	memset(&m_d3dmaterial,0, sizeof(m_d3dmaterial));
	memset(&m_pRamp, 0, sizeof(m_pRamp));
	memset(&m_viewport,0, sizeof(m_viewport));
	memset(m_pLight, 0, sizeof(D3DLIGHT8) * 16);
	memset(m_bLight, 0, sizeof(BOOL) * 16);
	memset(m_pPlane, 0, sizeof(float) * 32);
	memset(&m_pClipStatus, 0, sizeof(m_pClipStatus));

	m_PaletteNumber = 0;

	m_pVertexShaderHandle = 0;

	memset(&m_MatView,0,sizeof(m_MatView));
	m_MatView._11 = m_MatView._22 = m_MatView._33 = m_MatView._44 = 1.f;

	memset(&m_PrjView,0,sizeof(m_PrjView));
	m_PrjView._11 = m_PrjView._22 = m_PrjView._33 = m_PrjView._44 = 1.f;

	memset(&m_WldView,0,sizeof(m_WldView));
	m_WldView._11 = m_WldView._22 = m_WldView._33 = m_WldView._44 = 1.f;

	memset(m_pTexture,0,8 * sizeof(void*));
	m_pZStencilSurface = 0;
	m_pBackBuffer = 0;

	p->GetAdapterDisplayMode(0, &m_pDisplayMode);
	GL_GetCreationParameters(((BIDirect3D8*)p)->m_pGlDevice, &m_pCreationParameters);

	
	
}

ULONG
BIDirect3DDevice8::AddRef()
{
	APIDEBUG(AddRef);
	m_refCount++;
	return D3D_OK;	
}

HRESULT
BIDirect3DDevice8::ApplyStateBlock(DWORD Token)
{
	APIDEBUG(ApplyStateBlock);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::BeginScene()
{
	APIDEBUG(BeginScene);
	// Not needed in OpenGL
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::BeginStateBlock()
{
	APIDEBUG(BeginStateBlock);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::CaptureStateBlock(DWORD Token)
{
	APIDEBUG(CaptureStateBlock);
	return D3D_OK;
}

#define GETD3DCOLOR_BLUE(x)   (((x)&0xff) / 255.f)
#define GETD3DCOLOR_GREEN(x) ((((x)>>8)&0xff) / 255.f)
#define GETD3DCOLOR_RED(x)  ((((x)>>16)&0xff) / 255.f)
#define GETD3DCOLOR_ALPHA(x)  ((((x)>>24)&0xff) / 255.f)

HRESULT
BIDirect3DDevice8::Clear(DWORD Count, CONST D3DRECT *pRects, DWORD Flags, D3DCOLOR Color, float Z, DWORD Stencil)
{
	APIDEBUG(Clear);
	glClearColor(GETD3DCOLOR_RED(Color), GETD3DCOLOR_GREEN(Color), GETD3DCOLOR_BLUE(Color), GETD3DCOLOR_ALPHA(Color));
	glClearDepth(Z);
	glClearStencil(Stencil);

	// TODO: Implemente GL_WIN_swap_hint
	glClear(  ((Flags&D3DCLEAR_TARGET) ? GL_COLOR_BUFFER_BIT : 0)
			| ((Flags&D3DCLEAR_ZBUFFER) ? GL_DEPTH_BUFFER_BIT : 0)
			| ((Flags&D3DCLEAR_STENCIL) ? GL_STENCIL_BUFFER_BIT : 0)
			);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::CopyRects(IDirect3DSurface8 *pSourceSurface, CONST RECT *pSourceRectsArray, UINT cRects, IDirect3DSurface8 *pDestinationSurface, CONST POINT *pDestPointsArray)
{
	APIDEBUG(CopyRects);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::CreateAdditionalSwapChain(D3DPRESENT_PARAMETERS *pPresentationParameters, IDirect3DSwapChain8 **pSwapChain)
{
	APIDEBUG(CreateAdditionalSwapChain);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::CreateCubeTexture(UINT EdgeLength, UINT Levels, DWORD Usage, D3DFORMAT Format, D3DPOOL Pool, IDirect3DCubeTexture8 **ppCubeTexture)
{
	APIDEBUG(CreateCubeTexture);
	*ppCubeTexture = new BIDirect3DCubeTexture8(this, EdgeLength, Levels, Usage, Format, Pool);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::CreateDepthStencilSurface(UINT Width, UINT Height, D3DFORMAT Format, D3DMULTISAMPLE_TYPE MultiSample, IDirect3DSurface8 **ppSurface)
{
	APIDEBUG(CreateDepthStencilSurface);
	*ppSurface = new BIDirect3DSurface8(this, Width, Height, Format, MultiSample);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::CreateImageSurface(UINT Width, UINT Height, D3DFORMAT Format, IDirect3DSurface8 **ppSurface)
{
	APIDEBUG(CreateImageSurface);
	*ppSurface = new BIDirect3DSurface8(this, Width, Height, Format);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::CreateIndexBuffer(UINT Length, DWORD Usage, D3DFORMAT Format, D3DPOOL Pool, IDirect3DIndexBuffer8 **ppIndexBuffer)
{
	APIDEBUG(CreateIndexBuffer);
	*ppIndexBuffer = new BIDirect3DIndexBuffer8(this, Length, Usage, Format, Pool);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::CreatePixelShader(CONST DWORD *pFunction, DWORD *pHandle)
{
	APIDEBUG(CreatePixelShader);
	*pHandle = 0;
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::CreateRenderTarget(UINT Width, UINT Height, D3DFORMAT Format, D3DMULTISAMPLE_TYPE MultiSample, BOOL Lockable, IDirect3DSurface8 **ppSurface)
{
	APIDEBUG(CreateRenderTarget);
	*ppSurface = new BIDirect3DSurface8(this, Width, Height, Format, MultiSample, Lockable);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::CreateStateBlock(D3DSTATEBLOCKTYPE Type, DWORD *pToken)
{
	APIDEBUG(CreateStateBlock);
	*pToken = 0;
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::CreateTexture(UINT Width, UINT Height, UINT Levels, DWORD Usage, D3DFORMAT Format, D3DPOOL Pool, IDirect3DTexture8 **ppTexture)
{
	APIDEBUG(CreateTexture);

	ASSERT(Width);
	ASSERT(Height);

	if (ppTexture)
	{
		*ppTexture = new BIDirect3DTexture8(this, Width, Height, Levels, Usage, Format, Pool);
	}	
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::CreateVertexBuffer(UINT Length, DWORD Usage, DWORD FVF, D3DPOOL pool, IDirect3DVertexBuffer8 **ppVertexBuffer)
{
	APIDEBUG(CreateVertexBuffer);
	*ppVertexBuffer = new BIDirect3DVertexBuffer8(this, Length, Usage, FVF, pool);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::CreateVertexShader(CONST DWORD *pDeclaration, CONST DWORD *pFunction, DWORD *pHandle, DWORD Usage)
{
	APIDEBUG(CreateVertexShader);
	*pHandle = 0;
	/*
		See : glGenProgramsNV, glLoadProgramNV, glBindProgramNV
		and : GL_EXT_VERTEX_SHADER
	*/

	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::CreateVolumeTexture(UINT Width, UINT Height, UINT Depth, UINT Levels, DWORD Usage, D3DFORMAT Format, D3DPOOL Pool, IDirect3DVolumeTexture8 **ppVolumeTexture)
{
	APIDEBUG(CreateVolumeTexture);
	*ppVolumeTexture = new BIDirect3DVolumeTexture8(this, Width, Height, Depth, Levels, Usage, Format, Pool);	
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::DeletePatch(UINT Handle)
{
	APIDEBUG(DeletePatch);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::DeletePixelShader(DWORD Handle)
{
	APIDEBUG(DeletePixelShader);
	return D3D_OK;
}


HRESULT
BIDirect3DDevice8::DrawIndexedPrimitive(D3DPRIMITIVETYPE type, UINT minIndex, UINT NumVertices, UINT startIndex, UINT PrimitiveCount)
{		
	APIDEBUG(DrawIndexedPrimitive);
	D3D8_DrawIndexedBufferToGL(type, GetCurrentFVF(), GetCurrentVertexData(), PrimitiveCount, (CONST BYTE*)GetCurrentIndexData() + startIndex,0);	
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::DrawIndexedPrimitiveUP(D3DPRIMITIVETYPE type, UINT MinVertexIndex, UINT NumVertexIndices, UINT PrimitiveCount, CONST void *pIndexData, D3DFORMAT IndexDataFormat, CONST void *pVertexStreamZeroData, UINT VertexStreamZeroStride)
{
	APIDEBUG(DrawIndexedPrimitiveUP);
	D3D8_DrawIndexedBufferToGL(type, GetCurrentFVF(), GetCurrentVertexData(), PrimitiveCount, pIndexData, VertexStreamZeroStride);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::DrawPrimitive(D3DPRIMITIVETYPE PrimitiveType, UINT StartVertex, UINT PrimitiveCount)
{
	APIDEBUG(DrawPrimitive);

	const char *pOpCodes = (const char *)GetCurrentVertexData();
	int VertexCount = PrimitiveCount * 3;

	glBegin(D3D8_GlEnum(PrimitiveType));
	int fvf = GetCurrentFVF();

	while (VertexCount)
	{		 
		pOpCodes = D3D8_DecodeToGLInstruction(pOpCodes, fvf);
		VertexCount--;
	}
	glEnd();
	return D3D_OK;
}


HRESULT
BIDirect3DDevice8::DrawPrimitiveUP(D3DPRIMITIVETYPE PrimitiveType, UINT PrimitiveCount, CONST void *pVertexStreamZeroData, UINT VertexStreamZeroStride)
{
	APIDEBUG(DrawPrimitiveUP);
	const char *pOpCodes = (const char *)pVertexStreamZeroData;
	glBegin(D3D8_GlEnum(PrimitiveType));
	while (PrimitiveCount)
	{
		D3D8_DecodeToGLInstruction(pOpCodes, GetCurrentFVF());
		pOpCodes+=VertexStreamZeroStride + GetCurrentFVFSize();
		PrimitiveCount--;
	}
	glEnd();
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::DrawRectPatch(UINT Handle, CONST float *pNumSegs, CONST D3DRECTPATCH_INFO *pRectPatchInfo)
{
	APIDEBUG(DrawRectPatch);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::DrawTriPatch(UINT Handle, CONST float *pNumSegs, CONST D3DTRIPATCH_INFO *pTriPatchInfo)
{
	APIDEBUG(DrawTriPatch);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::EndScene()
{
	// Not needed in OpenGL
	APIDEBUG(EndScene);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::EndStateBlock(DWORD *pToken)
{
	APIDEBUG(EndStateBlock);
	return D3D_OK;
}

UINT
BIDirect3DDevice8::GetAvailableTextureMem()
{
	APIDEBUG(GetAvailableTextureMem);
	return 1<<31;
}

HRESULT
BIDirect3DDevice8::GetBackBuffer(UINT BackBuffer, D3DBACKBUFFER_TYPE Type, IDirect3DSurface8 **ppBackBuffer)
{
	APIDEBUG(GetBackBuffer);
	*ppBackBuffer = m_pBackBuffer;
	return D3D_OK;
}


HRESULT
BIDirect3DDevice8::GetClipPlane(DWORD Index, float *pPlane)
{
	APIDEBUG(GetClipPlane);
	*pPlane = m_pPlane[Index];
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::GetClipStatus(D3DCLIPSTATUS8 *pClipStatus)
{	
	APIDEBUG(GetClipStatus);
	*pClipStatus = m_pClipStatus;
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::GetCreationParameters(D3DDEVICE_CREATION_PARAMETERS *pParameters)
{
	APIDEBUG(GetCreationParameters);
	*pParameters = m_pCreationParameters;


	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::GetCurrentTexturePalette(UINT *PaletteNumber)
{
	APIDEBUG(GetCurrentTexturePalette);
	m_PaletteNumber = *PaletteNumber;
	return D3D_OK;
}



HRESULT
BIDirect3DDevice8::GetDepthStencilSurface(IDirect3DSurface8 **ppZStencilSurface)
{
	APIDEBUG(GetDepthStencilSurface);
	*ppZStencilSurface = new BIDirect3DSurface8(0,m_pDisplayMode.Width, m_pDisplayMode.Height,D3DFMT_D24S8,D3DMULTISAMPLE_NONE,0); // todo
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::GetDeviceCaps(D3DCAPS8 *pCaps)
{
	APIDEBUG(GetDeviceCaps);
	GLint n;
	GLfloat r[2];

	APIDEBUG(GetDeviceCaps);
	if (!pCaps)
		return D3DERR_INVALIDCALL;
	memset(pCaps, 0, sizeof(D3DCAPS8));

	   /* Device Info */
    pCaps->DeviceType = D3DDEVTYPE_HAL;
    pCaps->AdapterOrdinal = 0;

    /* Caps from DX7 Draw */
    pCaps->Caps = 0; //D3DCAPS_READ_SCANLINE;
    pCaps->Caps2 = D3DCAPS2_CANRENDERWINDOWED ;
    pCaps->Caps3 = 0;
    pCaps->PresentationIntervals = D3DPRESENT_INTERVAL_IMMEDIATE; // see wglSwapIntervalEXT

    /* Cursor Caps */
    pCaps->CursorCaps = 0;

    /* 3D Device Caps */
    pCaps->DevCaps = D3DDEVCAPS_EXECUTESYSTEMMEMORY
					|D3DDEVCAPS_EXECUTEVIDEOMEMORY
					|D3DDEVCAPS_TLVERTEXSYSTEMMEMORY
					|D3DDEVCAPS_DRAWPRIMTLVERTEX
					|D3DDEVCAPS_CANRENDERAFTERFLIP
					|D3DDEVCAPS_TEXTURENONLOCALVIDMEM
					|D3DDEVCAPS_DRAWPRIMITIVES2
					|D3DDEVCAPS_DRAWPRIMITIVES2EX
					|D3DDEVCAPS_HWTRANSFORMANDLIGHT
					|D3DDEVCAPS_HWRASTERIZATION
					|D3DDEVCAPS_PUREDEVICE					
					;

    pCaps->PrimitiveMiscCaps =
					D3DPMISCCAPS_MASKZ|
					D3DPMISCCAPS_CULLNONE|
					D3DPMISCCAPS_CULLCW|
					D3DPMISCCAPS_CULLCCW|
					D3DPMISCCAPS_COLORWRITEENABLE|
					D3DPMISCCAPS_TSSARGTEMP|
					D3DPMISCCAPS_BLENDOP					
					;

    pCaps->RasterCaps = 
					D3DPRASTERCAPS_DITHER|
					D3DPRASTERCAPS_FOGVERTEX|
					D3DPRASTERCAPS_FOGTABLE|
					D3DPRASTERCAPS_MIPMAPLODBIAS|
					D3DPRASTERCAPS_ZBIAS|
					D3DPRASTERCAPS_FOGRANGE|
					D3DPRASTERCAPS_ANISOTROPY|	// Check extensions
					D3DPRASTERCAPS_WFOG|		
					D3DPRASTERCAPS_ZFOG					
					;




    pCaps->ZCmpCaps = 
					D3DPCMPCAPS_NEVER|
					D3DPCMPCAPS_LESS|
					D3DPCMPCAPS_EQUAL|
					D3DPCMPCAPS_LESSEQUAL|
					D3DPCMPCAPS_GREATER|
					D3DPCMPCAPS_NOTEQUAL|
					D3DPCMPCAPS_GREATEREQUAL|
					D3DPCMPCAPS_ALWAYS
					;


    pCaps->SrcBlendCaps = 
					D3DPBLENDCAPS_ZERO|
					D3DPBLENDCAPS_ONE|
					D3DPBLENDCAPS_SRCCOLOR|
					D3DPBLENDCAPS_INVSRCCOLOR|
					D3DPBLENDCAPS_SRCALPHA|
					D3DPBLENDCAPS_INVSRCALPHA|
					D3DPBLENDCAPS_INVDESTALPHA|
					D3DPBLENDCAPS_DESTCOLOR|
					D3DPBLENDCAPS_INVDESTCOLOR|
					D3DPBLENDCAPS_SRCALPHASAT|
					D3DPBLENDCAPS_BOTHSRCALPHA|
					D3DPBLENDCAPS_BOTHINVSRCALPHA
					;

    pCaps->DestBlendCaps = 
					D3DPBLENDCAPS_ZERO|
					D3DPBLENDCAPS_ONE|
					D3DPBLENDCAPS_SRCCOLOR|
					D3DPBLENDCAPS_INVSRCCOLOR|
					D3DPBLENDCAPS_SRCALPHA|
					D3DPBLENDCAPS_INVSRCALPHA|
					D3DPBLENDCAPS_INVDESTALPHA|
					D3DPBLENDCAPS_DESTCOLOR|
					D3DPBLENDCAPS_INVDESTCOLOR|
					D3DPBLENDCAPS_SRCALPHASAT|
					D3DPBLENDCAPS_BOTHSRCALPHA|
					D3DPBLENDCAPS_BOTHINVSRCALPHA
					;
    pCaps->AlphaCmpCaps = 
					D3DPCMPCAPS_NEVER|
					D3DPCMPCAPS_LESS|
					D3DPCMPCAPS_EQUAL|
					D3DPCMPCAPS_LESSEQUAL|
					D3DPCMPCAPS_GREATER|
					D3DPCMPCAPS_NOTEQUAL|
					D3DPCMPCAPS_GREATEREQUAL|
					D3DPCMPCAPS_ALWAYS
					;

    pCaps->ShadeCaps = 
					D3DPSHADECAPS_COLORGOURAUDRGB|
					D3DPSHADECAPS_SPECULARGOURAUDRGB|
					D3DPSHADECAPS_ALPHAGOURAUDBLEND|
					D3DPSHADECAPS_FOGGOURAUD
					;

    pCaps->TextureCaps = 
					D3DPTEXTURECAPS_PERSPECTIVE|
					D3DPTEXTURECAPS_POW2|
					D3DPTEXTURECAPS_ALPHA|
					D3DPTEXTURECAPS_ALPHAPALETTE|
					D3DPTEXTURECAPS_PROJECTED|
					D3DPTEXTURECAPS_CUBEMAP|		// Check extension
					D3DPTEXTURECAPS_MIPMAP|		// Check extension
					D3DPTEXTURECAPS_MIPCUBEMAP|	// Check extension
					D3DPTEXTURECAPS_CUBEMAP_POW2// Check extension					
					;		
    pCaps->TextureFilterCaps =           // D3DPTFILTERCAPS for IDirect3DTexture8's
					D3DPTFILTERCAPS_MINFPOINT|
					D3DPTFILTERCAPS_MINFLINEAR|
					D3DPTFILTERCAPS_MINFANISOTROPIC|	// Check
					D3DPTFILTERCAPS_MIPFPOINT|
					D3DPTFILTERCAPS_MIPFLINEAR|
					D3DPTFILTERCAPS_MAGFPOINT|
					D3DPTFILTERCAPS_MAGFLINEAR
					;

		
    pCaps->CubeTextureFilterCaps = 0;      // D3DPTFILTERCAPS for IDirect3DCubeTexture8's
    pCaps->VolumeTextureFilterCaps = 0;    // D3DPTFILTERCAPS for IDirect3DVolumeTexture8's
    pCaps->TextureAddressCaps =		       // D3DPTADDRESSCAPS for IDirect3DTexture8's
					D3DPTADDRESSCAPS_WRAP|
					D3DPTADDRESSCAPS_MIRROR|	// check extension
					D3DPTADDRESSCAPS_CLAMP|	// check extension
					D3DPTADDRESSCAPS_INDEPENDENTUV	// check extension
					;
			
    pCaps->VolumeTextureAddressCaps = 0;   // D3DPTADDRESSCAPS for IDirect3DVolumeTexture8's

    pCaps->LineCaps =                    // D3DLINECAPS
					D3DLINECAPS_TEXTURE|
					D3DLINECAPS_ZTEST|
					D3DLINECAPS_BLEND|
					D3DLINECAPS_ALPHACMP|
					D3DLINECAPS_FOG
					;

	
	glGetIntegerv(GL_MAX_TEXTURE_SIZE, &n);

    pCaps->MaxTextureWidth = n; 
	pCaps->MaxTextureHeight = n;			// would explain d3dx retained mode crash ?
    pCaps->MaxVolumeExtent = 0;

    pCaps->MaxTextureRepeat = 8192;			// Geforce 2MX
    pCaps->MaxTextureAspectRatio = 0;		// ?

#if GL_EXT_texture_filter_anisotropic
	glGetIntegerv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &n);
    pCaps->MaxAnisotropy = n;
#else
	pCaps->MaxAnisotropy = 0;
#endif

    pCaps->MaxVertexW = 1E+010;

    pCaps->GuardBandLeft = -1E+008;
    pCaps->GuardBandTop = -1E+008;
    pCaps->GuardBandRight = 1E+008;
    pCaps->GuardBandBottom = 1E+008;

    pCaps->ExtentsAdjust = 0;
    pCaps->StencilCaps = 
					D3DSTENCILCAPS_KEEP|
					D3DSTENCILCAPS_ZERO|
					D3DSTENCILCAPS_REPLACE|
					D3DSTENCILCAPS_INCRSAT|	// check extension
					D3DSTENCILCAPS_DECRSAT|	// check extension
					D3DSTENCILCAPS_INVERT|	// check extension
					D3DSTENCILCAPS_INCR|	// check extension
					D3DSTENCILCAPS_DECR	// check extension
					;
    pCaps->FVFCaps = D3DFVFCAPS_DONOTSTRIPELEMENTS;
    pCaps->TextureOpCaps = 
					D3DTEXOPCAPS_DISABLE|
					D3DTEXOPCAPS_SELECTARG1|	// check extension
					D3DTEXOPCAPS_SELECTARG2|	// check extension
					D3DTEXOPCAPS_MODULATE|		// check extension
					D3DTEXOPCAPS_MODULATE2X|	// check extension
					D3DTEXOPCAPS_MODULATE4X|	// check extension
					D3DTEXOPCAPS_ADD|			// check extension
					D3DTEXOPCAPS_ADDSIGNED|	// check extension
					D3DTEXOPCAPS_ADDSIGNED2X|	// check extension
					D3DTEXOPCAPS_SUBTRACT|	// check extension
					D3DTEXOPCAPS_ADDSMOOTH|	// check extension
					D3DTEXOPCAPS_ADDSIGNED|	// check extension
					D3DTEXOPCAPS_BLENDDIFFUSEALPHA|	// check extension
					D3DTEXOPCAPS_BLENDTEXTUREALPHA|	// check extension
					D3DTEXOPCAPS_BLENDFACTORALPHA|	// check extension
					D3DTEXOPCAPS_BLENDTEXTUREALPHAPM|	// check extension
					D3DTEXOPCAPS_BLENDCURRENTALPHA|	// check extension
					D3DTEXOPCAPS_MODULATEALPHA_ADDCOLOR|	// check extension
					D3DTEXOPCAPS_MODULATECOLOR_ADDALPHA|	// check extension
					D3DTEXOPCAPS_MODULATEINVALPHA_ADDCOLOR|	// check extension
					D3DTEXOPCAPS_MODULATEINVCOLOR_ADDALPHA|	// check extension
					D3DTEXOPCAPS_DOTPRODUCT3|	// check extension
					D3DTEXOPCAPS_MULTIPLYADD|	// check extension
					D3DTEXOPCAPS_LERP	// check extension
					;


    pCaps->MaxTextureBlendStages = 8;

#if GL_ARB_multitexture
	glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB, &n);
    pCaps->MaxSimultaneousTextures = n;
#else
	pCaps->MaxSimultaneousTextures = 1;
#endif

    pCaps->VertexProcessingCaps = 
					D3DVTXPCAPS_DIRECTIONALLIGHTS|
					D3DVTXPCAPS_LOCALVIEWER|
					D3DVTXPCAPS_POSITIONALLIGHTS|
					D3DVTXPCAPS_TEXGEN
					;	
	
	glGetIntegerv(GL_MAX_LIGHTS, &n);
    pCaps->MaxActiveLights = n;

	glGetIntegerv(GL_MAX_CLIP_PLANES, &n);
    pCaps->MaxUserClipPlanes = n;

	n = 0;
#if GL_ARB_vertex_blend
	glGetIntegerv(GL_MAX_VERTEX_UNITS_ARB, &n);	
#endif

	pCaps->MaxVertexBlendMatrices = n;		//	GL_MAX_VERTEX_UNITS_ARB ?
    pCaps->MaxVertexBlendMatrixIndex = 0;
	
	glGetFloatv(GL_POINT_SIZE_RANGE, r);
    pCaps->MaxPointSize = r[1];

    pCaps->MaxPrimitiveCount = 65535;       // max number of primitives per DrawPrimitive call
    pCaps->MaxVertexIndex = 65535;			// 16bit max for now
    pCaps->MaxStreams = 0 ;					// Direct3D 7 like
    pCaps->MaxStreamStride = 65535;			// max stride for SetStreamSource

    pCaps->VertexShaderVersion = 0;
    pCaps->MaxVertexShaderConst = 0;		// number of vertex shader constant registers

    pCaps->PixelShaderVersion = 0;
    pCaps->MaxPixelShaderValue = 0;        // max value of pixel shader arithmetic component

							
	return D3D_OK;
}

/*

IDirect3DDevice8::GetDirect3D
Returns an interface to the instance of the Microsoft Direct3D object that created the device.

HRESULT GetDirect3D(
  IDirect3D8** ppD3D8
);

Parameters 

  ppD3D8 
	[out, retval] Address of a pointer to an IDirect3D8 interface, representing the interface of the Direct3D object that created the device. 

  Return Values
	If the method succeeds, the return value is D3D_OK.

	If the method fails, the return value can be D3DERR_INVALIDCALL.

Remarks
	Calling GetDirect3D increases the reference count on the interface, so applications should call IUnknown::Release through this pointer when finished with the interface.
	
Note  
	Calling this method will increase the internal reference count on the IDirect3D8 interface. Failure to call IUnknown::Release when finished using this IDirect3D8 interface results in a memory leak.

*/

HRESULT
BIDirect3DDevice8::GetDirect3D(IDirect3D8 **ppD3D)
{

	APIDEBUG(GetDirect3D);
	m_pIDirect3D8->AddRef();
	*ppD3D = m_pIDirect3D8;

	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::GetDisplayMode(D3DDISPLAYMODE *pMode)
{
	APIDEBUG(GetDisplayMode);
	*pMode = m_pDisplayMode;
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::GetFrontBuffer(IDirect3DSurface8 *pDestSurface)
{	
	APIDEBUG(GetFrontBuffer);
	return D3D_OK;
}

void
BIDirect3DDevice8::GetGammaRamp(D3DGAMMARAMP *pRamp)
{
	APIDEBUG(GetGammaRamp);
	*pRamp = m_pRamp;
	return ;
}

HRESULT
BIDirect3DDevice8::GetIndices(IDirect3DIndexBuffer8 **ppIndexData, UINT *pBaseVertexIndex)
{	
	APIDEBUG(GetIndices);
	*ppIndexData = 0;
	*pBaseVertexIndex = 0;	
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::GetInfo(DWORD DevInfoID, void *pDevInfoStruct, DWORD DevInfoStructSize)
{	
	APIDEBUG(GetInfo);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::GetLight(DWORD Index, D3DLIGHT8 *p)
{	
	APIDEBUG(GetLight);
	*p = m_pLight[Index];
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::GetLightEnable(DWORD Index , BOOL *pEnable)
{
	APIDEBUG(GetLightEnable);
	*pEnable = m_bLight[Index];
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::GetMaterial(D3DMATERIAL8 *pMaterial)
{
	APIDEBUG(GetMaterial);
	*pMaterial = m_d3dmaterial;
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::GetPaletteEntries(UINT PaletteNumber, PALETTEENTRY *pEntries)
{
	APIDEBUG(GetPaletteEntries);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::GetPixelShader(DWORD *pHandle)
{
	APIDEBUG(GetPixelShader);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::GetPixelShaderConstant(DWORD Register, void *pConstantData, DWORD ConstantCount)
{
	APIDEBUG(GetPixelShaderConstant);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::GetRasterStatus(D3DRASTER_STATUS *pRasterStatus)
{
	APIDEBUG(GetRasterStatus);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::GetRenderState(D3DRENDERSTATETYPE State, DWORD *pValue)
{
	APIDEBUG(GetRenderState);
	*pValue  = m_RenderState[State];
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::GetRenderTarget(IDirect3DSurface8 **ppRenderTarget)
{
	APIDEBUG(GetRenderTarget);
	*ppRenderTarget = new BIDirect3DSurface8(this, m_pDisplayMode.Width, m_pDisplayMode.Height, m_pDisplayMode.Format, D3DMULTISAMPLE_NONE,TRUE);	// TODO!
	return D3D_OK;
}


HRESULT
BIDirect3DDevice8::GetStreamSource(UINT StreamNumber, IDirect3DVertexBuffer8 **ppStreamData, UINT *pStride)
{
	APIDEBUG(GetStreamSource);
	*ppStreamData = 0;
	return D3DERR_INVALIDCALL;
}

HRESULT
BIDirect3DDevice8::GetTexture(DWORD Stage, IDirect3DBaseTexture8 **ppTexture)
{
	APIDEBUG(GetTexture);
	*ppTexture = m_pTexture[Stage];
	
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::GetTextureStageState(DWORD Stage, D3DTEXTURESTAGESTATETYPE Type, DWORD *pValue)
{
	APIDEBUG(GetTextureStageState);
	*pValue = 0;
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::GetTransform(D3DTRANSFORMSTATETYPE State, D3DMATRIX *pMatrix)
{
	APIDEBUG(GetTransform);
	switch(State)
	{
		case D3DTS_WORLD:
			*pMatrix = m_WldView;
		break;
		case D3DTS_VIEW:
			*pMatrix = m_MatView;
		break;

		case D3DTS_PROJECTION:
			*pMatrix = m_PrjView;
		break;        
		default:
		break;
	}
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::GetVertexShader(DWORD *pHandle)
{
	APIDEBUG(GetVertexShader);
	*pHandle = m_pVertexShaderHandle;
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::GetVertexShaderConstant(DWORD Register, void *pConstantData, DWORD ConstantCount)
{
	APIDEBUG(GetVertexShaderConstant);
	return D3DERR_INVALIDCALL;
}

HRESULT
BIDirect3DDevice8::GetVertexShaderDeclaration(DWORD Handle, void *pData, DWORD *pSizeOfData)
{
	APIDEBUG(GetVertexShaderDeclaration);
	return D3DERR_INVALIDCALL;
}

HRESULT
BIDirect3DDevice8::GetVertexShaderFunction(DWORD Handle, void *pData, DWORD *pSizeOfData)
{
	APIDEBUG(GetVertexShaderDeclaration);
	return D3DERR_INVALIDCALL;
}

HRESULT
BIDirect3DDevice8::GetViewport(D3DVIEWPORT8 *pViewport)
{
	APIDEBUG(GetViewport);
	*pViewport = m_viewport;
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::LightEnable(DWORD Index, BOOL Enable)
{
	APIDEBUG(LightEnable);
	m_bLight[Index] = Enable;

	if (Enable)
		glEnable(GL_LIGHT0 + Index);
	else
		glDisable(GL_LIGHT0 + Index);
	
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::MultiplyTransform(D3DTRANSFORMSTATETYPE State, CONST D3DMATRIX *pMatrix)
{
	APIDEBUG(MultiplyTransform);
	switch(State)
	{
		case D3DTS_WORLD:
			
		break;
		case D3DTS_VIEW:
			glMatrixMode( GL_MODELVIEW );
			glMultMatrixf( (float*) pMatrix);			
		break;
		case D3DTS_PROJECTION:
			glMatrixMode( GL_PROJECTION );
			glMultMatrixf( (float*) pMatrix);	//todo : LHtoRH		
		break;        
		default:
		break;
	}
	
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::Present(CONST RECT *pSourceRect, CONST RECT *pDestRect, HWND hDestWindowOverride, CONST RGNDATA *pDirtyRegion)
{
	APIDEBUG(Present);
	// TODO : Page flipping
	BIDirect3D8 * p = (BIDirect3D8*)m_pIDirect3D8;
	
	GL_PresentDevice(p->m_pGlDevice);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::ProcessVertices(UINT SrcStartIndex, UINT DestIndex, UINT VertexCount, IDirect3DVertexBuffer8 *pDestBuffer, DWORD Flags)
{
	APIDEBUG(ProcessVertices);
	return D3DERR_INVALIDCALL;
}

HRESULT
BIDirect3DDevice8::QueryInterface(REFIID riid, void **ppvObj)
{
	APIDEBUG(QueryInterface);
	*ppvObj = 0;
	return D3DERR_INVALIDCALL;
}

ULONG
BIDirect3DDevice8::Release()
{
	m_refCount--;
	if (!m_refCount)
	{
		APIDEBUG(RELEASEDDRIVER);
		delete this;
	}
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::Reset(D3DPRESENT_PARAMETERS *pPresentationParameters)
{	
	APIDEBUG(Reset);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::ResourceManagerDiscardBytes(DWORD Bytes)
{
	APIDEBUG(ResourceManagerDiscardBytes);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::SetClipPlane(DWORD Index, CONST float *pPlane)
{
	APIDEBUG(SetClipPlane);
	m_pPlane[Index] = *pPlane;
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::SetClipStatus(CONST D3DCLIPSTATUS8 *pClipStatus)
{
	APIDEBUG(SetClipStatus);
	m_pClipStatus = *pClipStatus;
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::SetCurrentTexturePalette(UINT PaletteNumber)
{
	APIDEBUG(SetCurrentTexturePalette);
	m_PaletteNumber = PaletteNumber;
	// TODO: Write palette mgr
	return D3D_OK;
}

void
BIDirect3DDevice8::SetCursorPosition(UINT XScreenSpace, UINT YScreenSpace, DWORD Flags)
{
	APIDEBUG(SetCursorPosition);
	// TODO: Find equivalent extensions ? (SYSTEM dependant ?)
	return ;
}

HRESULT
BIDirect3DDevice8::SetCursorProperties(UINT XHotSpot, UINT YHotSport, IDirect3DSurface8 *pCursorBitmap)
{
	APIDEBUG(SetCursorProperties);
	// TODO: Find equivalent extensions ? (SYSTEM dependant ?)
	return D3D_OK;
}

void
BIDirect3DDevice8::SetGammaRamp(DWORD Flags, CONST D3DGAMMARAMP *pRamp)
{
	APIDEBUG(SetGammaRamp);
	m_pRamp = *pRamp;
	// TODO: Find equivalent extensions ?
	return ;
}

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif
#define RTOD(rad) (float)(rad * 180. / M_PI)

HRESULT
BIDirect3DDevice8::SetLight(DWORD Index, CONST D3DLIGHT8 *d3dlite)
{
	APIDEBUG(SetLight);
	m_pLight[Index] = *d3dlite;	

	int i = GL_LIGHT0 + Index;
	glLightfv( i, GL_AMBIENT, (float*)&d3dlite->Ambient);
	glLightfv( i, GL_DIFFUSE, (float*)&d3dlite->Diffuse);
	glLightfv( i, GL_SPECULAR, (float*)&d3dlite->Specular);
	glLightf( i, GL_CONSTANT_ATTENUATION, d3dlite->Attenuation0);
	glLightf( i, GL_LINEAR_ATTENUATION, d3dlite->Attenuation1);
	glLightf( i, GL_QUADRATIC_ATTENUATION, d3dlite->Attenuation2);
		
	float l[4];

	if (d3dlite->Type == D3DLIGHT_DIRECTIONAL)
	{
		l[0]=d3dlite->Direction.x;
		l[1]=d3dlite->Direction.y;
		l[2]=d3dlite->Direction.z;
		l[3]=0;
		glLightfv( i, GL_POSITION, l);
		glLightf( i, GL_SPOT_CUTOFF, 0);
		glLightf( i, GL_SPOT_EXPONENT, 180);
	
	}
	else if (d3dlite->Type == D3DLIGHT_SPOT)
	{
		l[0]=d3dlite->Position.x;
		l[1]=d3dlite->Position.y;
		l[2]=d3dlite->Position.z;
		l[3]=1;
		glLightfv( i, GL_POSITION, l);
		
		l[0]=d3dlite->Direction.x;
		l[1]=d3dlite->Direction.y;
		l[2]=d3dlite->Direction.z;
		glLightfv( i, GL_SPOT_DIRECTION, l);

		glLightf( i, GL_SPOT_CUTOFF, RTOD(d3dlite->Phi));
		glLightf( i, GL_SPOT_EXPONENT, RTOD(d3dlite->Falloff));
	
	}
	else
	{
		l[0]=d3dlite->Position.x;
		l[1]=d3dlite->Position.y;
		l[2]=d3dlite->Position.z;
		l[3]=1;
		glLightfv( i, GL_POSITION, l);
	}	
	return D3D_OK;
}


HRESULT
BIDirect3DDevice8::SetMaterial(CONST D3DMATERIAL8 *pMaterial)
{
	APIDEBUG(SetMaterial);
	m_d3dmaterial = *pMaterial;
	
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, (float*)&m_d3dmaterial.Emissive);
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, (float*)&m_d3dmaterial.Ambient);	
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, (float*)&m_d3dmaterial.Diffuse);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, (float*)&m_d3dmaterial.Specular);	
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, m_d3dmaterial.Power);

	// TODO : Set material
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::SetPaletteEntries(UINT PaletteNumber, CONST PALETTEENTRY *pEntries)
{
	APIDEBUG(SetPaletteEntries);
	// TODO: use glColorTable
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::SetPixelShader(DWORD pHandle)
{
	APIDEBUG(SetPixelShader);
	// TODO: Use NV or ATI shader
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::SetPixelShaderConstant(DWORD Register, CONST void *pConstantData, DWORD ConstantCount)
{
	APIDEBUG(SetPixelShaderConstant);
	// TODO: Use NV or ATI shader
	return D3D_OK;
}

static GLenum CmpFunc[] = {
	0,
	GL_NEVER,
    GL_LESS,
    GL_EQUAL,
    GL_LEQUAL,
    GL_GREATER,
    GL_NOTEQUAL,
    GL_GEQUAL,
    GL_ALWAYS
};

static GLenum FillMode[] = {
	0,
	GL_POINT,
	GL_LINE,
	GL_FILL
};

static GLenum StencilOp[] = {
	0,
	GL_KEEP, 
	GL_ZERO, 
	GL_REPLACE,
	GL_INCR,	// SAT
	GL_DECR,	// SAT
	GL_INVERT,
	GL_INCR,
	GL_DECR	
};


HRESULT
BIDirect3DDevice8::SetRenderState(D3DRENDERSTATETYPE State, DWORD pValue)
{
	APIDEBUG(SetRenderState);
	//  RenderState fast state
	if (m_RenderState[State] == pValue)
		return D3D_OK;

	float fValue = *(float*)&pValue;
	
	m_RenderState[State] = pValue;
	switch(State)
	{
		case D3DRS_ZENABLE					:	if (pValue) glEnable(GL_DEPTH); else glDisable(GL_DEPTH); break;
		case D3DRS_FILLMODE					:	glPolygonMode( GL_FRONT_AND_BACK, FillMode[pValue]); break;
		case D3DRS_SHADEMODE				:	glShadeModel( pValue == D3DSHADE_FLAT ? GL_FLAT : GL_SMOOTH ); break;
		case D3DRS_LINEPATTERN				:	break;
		case D3DRS_ZWRITEENABLE				:	glDepthMask( pValue ? GL_TRUE : GL_FALSE); break;
		case D3DRS_ALPHATESTENABLE			:	break;
		case D3DRS_LASTPIXEL				:	break;
		case D3DRS_SRCBLEND					:	
		case D3DRS_DESTBLEND				:	glBlendFunc(m_RenderState[D3DRS_SRCBLEND], m_RenderState[D3DRS_DESTBLEND]); break;
		case D3DRS_CULLMODE					:	
			switch(pValue)
			{
				case D3DCULL_NONE:
				glDisable(GL_CULL_FACE);
				break;
				case D3DCULL_CCW:
				glEnable(GL_CULL_FACE);
				glCullFace(GL_FRONT);
				break;
				case D3DCULL_CW:
				glEnable(GL_CULL_FACE);
				glCullFace(GL_BACK);
				break;
			}
		break;
		case D3DRS_ZFUNC					:	glDepthFunc( CmpFunc[pValue] ); break;
		case D3DRS_ALPHAREF					:	glAlphaFunc( CmpFunc[pValue], *(float*)&m_RenderState[D3DRS_ALPHAREF]);
		case D3DRS_ALPHAFUNC				:	glAlphaFunc( CmpFunc[pValue], *(float*)&m_RenderState[D3DRS_ALPHAREF]); break;
		case D3DRS_DITHERENABLE				:	if (pValue) glEnable(GL_DITHER); else glDisable(GL_DITHER); break;
		case D3DRS_ALPHABLENDENABLE			:	if (pValue) glEnable(GL_BLEND); else glDisable(GL_BLEND); break;
		case D3DRS_FOGENABLE				:	if (pValue) glEnable(GL_FOG); else glDisable(GL_FOG); break;
		case D3DRS_SPECULARENABLE			:	break;
		case D3DRS_ZVISIBLE					:	break;
		case D3DRS_FOGCOLOR					:	{ float v[3] = { float(pValue & 0xff)/255.f, 
																 float((pValue >> 8)&0xff)/255.f, 
																 float((pValue>>16)&0xff)/255.f }; glFogfv(GL_FOG_COLOR, v); } break; 
		case D3DRS_FOGTABLEMODE				:	break;
		case D3DRS_FOGSTART					:	glFogf(GL_FOG_START, fValue); break;
		case D3DRS_FOGEND					:	glFogf(GL_FOG_END, fValue); break;
		case D3DRS_FOGDENSITY				:	glFogf(GL_FOG_DENSITY, fValue); break;
		case D3DRS_EDGEANTIALIAS			:	break;
		case D3DRS_ZBIAS					:	break;
		case D3DRS_RANGEFOGENABLE			:	break;
		case D3DRS_STENCILENABLE			:	if (pValue) glEnable(GL_STENCIL_TEST); else glDisable(GL_STENCIL_TEST); 
												break;
		case D3DRS_STENCILFAIL				:	
		case D3DRS_STENCILZFAIL				:	
		case D3DRS_STENCILPASS				:	
		case D3DRS_STENCILFUNC				:	
		case D3DRS_STENCILREF				:	
		case D3DRS_STENCILMASK				:	
		case D3DRS_STENCILWRITEMASK			:	

		glStencilFunc(CmpFunc[m_RenderState[D3DRS_STENCILFUNC]], m_RenderState[D3DRS_STENCILREF], m_RenderState[D3DRS_STENCILMASK]);
		glStencilOp(StencilOp[m_RenderState[D3DRS_STENCILFAIL]], StencilOp[m_RenderState[D3DRS_STENCILPASS]], StencilOp[m_RenderState[D3DRS_STENCILZFAIL]]);
		
		break;


		case D3DRS_TEXTUREFACTOR			:	break;
		case D3DRS_WRAP0					:	break;
		case D3DRS_WRAP1					:	break;
		case D3DRS_WRAP2					:	break;
		case D3DRS_WRAP3					:	break;
		case D3DRS_WRAP4					:	break;
		case D3DRS_WRAP5					:	break;
		case D3DRS_WRAP6					:	break;
		case D3DRS_WRAP7					:	break;
		case D3DRS_CLIPPING					:	break;
		case D3DRS_LIGHTING					:	if (pValue) glEnable(GL_LIGHTING); else glDisable(GL_LIGHTING); break;
		case D3DRS_AMBIENT					:	break;
		case D3DRS_FOGVERTEXMODE			:	break;
		case D3DRS_COLORVERTEX				:	break;
		case D3DRS_LOCALVIEWER				:	break;
		case D3DRS_NORMALIZENORMALS			:	if (pValue) glEnable(GL_NORMALIZE); else glDisable(GL_NORMALIZE); break;
		case D3DRS_DIFFUSEMATERIALSOURCE	:	break;
		case D3DRS_SPECULARMATERIALSOURCE	:	break;
		case D3DRS_AMBIENTMATERIALSOURCE	:	break;
		case D3DRS_EMISSIVEMATERIALSOURCE	:	break;
		case D3DRS_VERTEXBLEND				:	break;
		case D3DRS_CLIPPLANEENABLE			:	break;
		case D3DRS_SOFTWAREVERTEXPROCESSING :	break;
		case D3DRS_POINTSIZE				:	break;
		case D3DRS_POINTSIZE_MIN			:	break;
		case D3DRS_POINTSPRITEENABLE		:	break;
		case D3DRS_POINTSCALEENABLE			:	break;
		case D3DRS_POINTSCALE_A				:	break;
		case D3DRS_POINTSCALE_B				:	break;
		case D3DRS_POINTSCALE_C				:	break;
		case D3DRS_MULTISAMPLEANTIALIAS		:	break;
		case D3DRS_MULTISAMPLEMASK			:	break;
		case D3DRS_PATCHEDGESTYLE			:	break;
		case D3DRS_PATCHSEGMENTS			:	break;
		case D3DRS_DEBUGMONITORTOKEN		:	break;
		case D3DRS_POINTSIZE_MAX			:	break;
		case D3DRS_INDEXEDVERTEXBLENDENABLE :	break;
		case D3DRS_COLORWRITEENABLE			:	break;
		case D3DRS_TWEENFACTOR				:	break;
		case D3DRS_BLENDOP					:	break;
	}
		
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::SetRenderTarget(IDirect3DSurface8 *pRenderTarget, IDirect3DSurface8 *pNewZStencil)
{
	APIDEBUG(SetRenderTarget);	// TODO!
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::SetStreamSource(UINT StreamNumber, IDirect3DVertexBuffer8 *ppStreamData, UINT pStride)
{
	APIDEBUG(SetStreamSource);
	m_pCurrentVertexBuffer = ppStreamData;
	m_nCurrentVertexStride = pStride;
	return D3D_OK;
}
HRESULT
BIDirect3DDevice8::SetTexture(DWORD Stage, IDirect3DBaseTexture8 *ppTexture)
{
	APIDEBUG(SetTexture);

	m_pTexture[Stage] = ppTexture;
	if (ppTexture)
		((BIDirect3DTexture8*)ppTexture)->BindTexture();
	else
		glDisable(GL_TEXTURE_2D);

	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::SetTextureStageState(DWORD Stage, D3DTEXTURESTAGESTATETYPE Type, DWORD pValue)
{
	APIDEBUG(SetTextureStageState);
	return D3D_OK;
}

/*

  This is one basic difference between OpenGL and Direct3D. 
  In both coordinate systems, the X and Y axis are as normal. 
  The Y positive axis going upwards and the X positive axis going to the right. 

  However, the two systems differ in their Z-axis implementation. 
  A Left-Handed System, which Direct3D uses, is one where the Z-axis is positive INTO the screen. 
  Take your Left Hand, face the palm upwards with your fingers pointing in the direction of the positive x-axis (to the right), and curl your fingers in the direction of the Y-positive axis (up). Now if you stick your thumb STRAIGHT (like you were giving a sideways "thumbs-up" sign), your thumb is now pointing in the direction of the positive Z-axis.
  
  OpenGL, on the other hand, uses a Right-Handed system, where the positive Z-axis is coming OUT of the screen. 
  Again, to remember this, take your right hand, point your palm upwards and your fingers pointing in the direction of the positive x-axis 
  (to the right). Now curl your fingers up to the direction of the positive y-axis (up), and point your thumb out. It is now pointing in the direction of the positive Z-axis (out)..

  Note: no doc exists how to do the conversion
*/


void prjLHtoRH(float *d, const float *s) // Exclusive function
{
	d[0] = -s[0];	d[1] = -s[1];	d[2] = -s[2];	d[3] = -s[3];
	d[4] = s[4];	d[5] = s[5];	d[6] = s[6];	d[7] = s[7];
	d[8] = -s[8];	d[9] = -s[9];	d[10] = -s[10];	d[11] = -s[11];
	d[12] = s[12];	d[13] = s[13];	d[14] = s[14];	d[15] = s[15];
}

void viewLHtoRH(float *d, const float *s)
{
	d[0] = -s[0];	d[1] = s[1];	d[2] = -s[2];	d[3] = s[3];
	d[4] = -s[4];	d[5] = s[5];	d[6] = -s[6];	d[7] = s[7];
	d[8] = -s[8];	d[9] = s[9];	d[10] = -s[10];	d[11] = s[11];
	d[12] = -s[12];	d[13] = s[13];	d[14] = -s[14];	d[15] = s[15];
}

HRESULT
BIDirect3DDevice8::SetTransform(D3DTRANSFORMSTATETYPE State, CONST D3DMATRIX *pMatrix)
{
	APIDEBUG(SetTransform);
	switch(State)
	{
		case D3DTS_TEXTURE0:
		glMatrixMode( GL_TEXTURE );
		glLoadMatrixf( (float*) pMatrix);
		break;

		case D3DTS_VIEW:
		{
			m_MatView = *pMatrix;
			glMatrixMode( GL_MODELVIEW );		
			float f[16];
			viewLHtoRH(f, (float*) &m_MatView);
			glLoadMatrixf(f);
			glMultMatrixf( (float*) &m_WldView);
		}
		
		break;

		case D3DTS_WORLD:
		{
			m_WldView = *pMatrix;
			glMatrixMode( GL_MODELVIEW );		
			float f[16];
			viewLHtoRH(f, (float*) &m_MatView);
			glLoadMatrixf(f);
			glMultMatrixf( (float*) &m_WldView);
		}
		
		break;

		case D3DTS_PROJECTION:
		m_PrjView = *pMatrix;
		glMatrixMode( GL_PROJECTION );
		float f[16];
		prjLHtoRH(f, (float*) &m_PrjView);
		glLoadMatrixf(f);	// TODO: D3D is RH and OpenGL is LH 
		

		break;        
		default:
		break;
	}
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::SetVertexShader(DWORD pHandle)
{
	APIDEBUG(SetVertexShader);
	m_pVertexShaderHandle = pHandle;	
	m_nCurrentStreamFVFSize = D3D8_DecodeLength(m_pVertexShaderHandle);	 
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::SetVertexShaderConstant(DWORD Register, CONST void *pConstantData, DWORD ConstantCount)
{
	APIDEBUG(SetVertexShaderConstant);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::SetViewport(CONST D3DVIEWPORT8 *pViewport)
{
	APIDEBUG(SetViewport);
	m_viewport = *pViewport;
	glViewport(m_viewport.X, m_viewport.Y, m_viewport.Width, m_viewport.Height);
	return D3D_OK;
}

BOOL
BIDirect3DDevice8::ShowCursor(BOOL bShow)
{
	APIDEBUG(ShowCursor);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::TestCooperativeLevel()
{
	APIDEBUG(TestCooperativeLevel);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::UpdateTexture(IDirect3DBaseTexture8 *pSourceTexture, IDirect3DBaseTexture8 *pDestinationTexture)
{
	APIDEBUG(UpdateTexture);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::ValidateDevice(DWORD *pNumPasses)
{
	APIDEBUG(ValidateDevice);
	*pNumPasses = 1;
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::DeleteStateBlock(DWORD Token)
{
	APIDEBUG(DeleteStateBlock);
	return D3D_OK;
}

HRESULT
BIDirect3DDevice8::DeleteVertexShader(DWORD Handle)
{
	APIDEBUG(DeleteVertexShader);
	return D3D_OK;
}

HRESULT 
BIDirect3DDevice8::SetIndices(IDirect3DIndexBuffer8* pIndexData, UINT BaseVertexIndex)
{
	APIDEBUG(SetIndices);
	m_pCurrentIndexBuffer = pIndexData;
	m_nBaseVertexIndex = BaseVertexIndex;
	return D3D_OK;
}

HRESULT 
BIDirect3DDevice8::GetPixelShaderFunction(DWORD Handle, void *pData, DWORD *pSizeOfData)
{
	APIDEBUG(GetPixelShaderFunction);
	return D3D_OK;
}
///

DWORD BIDirect3DDevice8::GetCurrentFVF() const
{
	APIDEBUG(GetCurrentFVF);
	return m_pVertexShaderHandle;
}

DWORD BIDirect3DDevice8::GetCurrentFVFSize() const
{
	APIDEBUG(GetCurrentFVFSize);
	return this->m_nCurrentStreamFVFSize;
}

CONST VOID *BIDirect3DDevice8::GetCurrentVertexData() const
{
	APIDEBUG(GetCurrentVertexData);
	return ((BIDirect3DVertexBuffer8*)this->m_pCurrentVertexBuffer)->m_pVtxBuffer;
}

CONST VOID *BIDirect3DDevice8::GetCurrentIndexData() const
{
	APIDEBUG(GetCurrentIndexData);
	return ((BIDirect3DIndexBuffer8*)this->m_pCurrentIndexBuffer)->m_pIndex;
}


